
#define TIM32BITMAX (67107840UL)                                        //Maximal 32-bit timer value (1024 * 65535 (= 0xFFFF)).

#include "StepMotorDriverA4988.h"
#include <avr/boot.h>                                                   //boot_lock_fuse_bits_get


const uint8_t MOTOR1POSLO = 0x01U;
const uint8_t MOTOR3POSLO = 0x02U;
const uint8_t MOTOR4POSLO = 0x04U;
const uint8_t MOTOR5POSLO = 0x08U;
const uint8_t MOTOR1POSHI = 0x10U;
const uint8_t MOTOR3POSHI = 0x20U;
const uint8_t MOTOR4POSHI = 0x40U;
const uint8_t MOTOR5POSHI = 0x80U;

uint8_t motorTimUpdateFlag;                                             //Bit 7 ... 4 are for stepmode Update (resp. timer 5, 4, 3, 1).
                                                                        //Bit 3 ... 0 are for timer output compare update.
uint8_t motorEndPos;                                                    //Bit 7 ... 4 are endpos enable; Bit 3 ... 0 indicate endpos has triggered.
uint8_t motorDir;                                                       //Bit 3 ... 0 are direction reminders.

#ifdef MOTOR_TIMER_1
  int32_t  motor1posCount;                                              //Counts the motor position in steps since the last reset.
  int32_t  motor1targetCount;                                           //Motor step end counter, when motorxEndPos bit 7...4 is set.
  uint8_t  motor1ocal;                                                  //New OCxAL register if motorTimUpdateFlag bit 3 ... 0 is set.
  uint8_t  motor1ocah;                                                  //New OCxAH register if motorTimUpdateFlag bit 3 ... 0 is set.
  MOTORSTEPSELECTION motor1micrStepMode;                                //Reminds the user settling for the step mode (Full, half,...).
  uint8_t  motor1micrStepBaseCount;                                     //Base value for motorxmicrStepCount to determine when a full step is reached.
  uint8_t  motor1micrStepCount;                                         //Counts actual motor (micro)steps.
#endif

#ifdef MOTOR_TIMER_3
  int32_t  motor3posCount;
  int32_t  motor3targetCount;
  uint8_t  motor3ocal;
  uint8_t  motor3ocah;
  MOTORSTEPSELECTION motor3micrStepMode;
  uint8_t  motor3micrStepBaseCount;
  uint8_t  motor3micrStepCount;
#endif

#ifdef MOTOR_TIMER_4
  int32_t  motor4posCount;
  int32_t  motor4targetCount;
  uint8_t  motor4ocal;
  uint8_t  motor4ocah;
  MOTORSTEPSELECTION motor4micrStepMode;
  uint8_t  motor4micrStepBaseCount;
  uint8_t  motor4micrStepCount;
#endif

#ifdef MOTOR_TIMER_5
  int32_t  motor5posCount;
  int32_t  motor5targetCount;
  uint8_t  motor5ocal;
  uint8_t  motor5ocah;
  MOTORSTEPSELECTION motor5micrStepMode;
  uint8_t  motor5micrStepBaseCount;
  uint8_t  motor5micrStepCount;
#endif







/********************************************************************************************************
 *    Read back a digital output state.
 *   [in]  pin      The Arduino digital pin number.
 *   [out] int      HIGH when output pin is '1', LOW in all other cases (pin not existing or pin LOW).
 */
static int
outputPinReadDigital (uint8_t pin)
{
  uint8_t bit = digitalPinToBitMask (pin);
  uint8_t port = digitalPinToPort (pin);
  if (port == NOT_A_PIN) return LOW;
  return (*portOutputRegister (port) & bit) ? HIGH : LOW;
}






/********************************************************************************************************
 *    Returns the IO clock count per second.
 *    The AVR core has an 8MHz RC or 128kHz system oscillator. These can be selected via programmable 'Fuses'.
 *    The external oscillator in Arduino is 16MHz and believed to be the major oscillator source. The function
 *    first determines the clock source used. Then, it checks the CLKPR (Clock Prescaler) register in case the
 *    user has reduced the speed to preserve energy. That results in a final amount of I/O clocks available to the application.
 *    [Out] uint32_t:  I/O clocks per second.
 */
uint32_t
motorTimerGetIOclocksPerSecond ()
{
  uint32_t ioClk = F_CPU;                                               //Expected 16MHz.
  switch (boot_lock_fuse_bits_get (0) & 0x0F)                           //Read the AVR clock programming fuses. Unless internal clock is active,
  {                                                                     //F_CPU is most likely the correct CPU clock frequency.
  case   3:     ioClk =  128000UL; break;                               //Running at 128kHz system clock.
  case   2:     ioClk = 8000000UL; break;                               //8MHz RC.
  default:      break;                                                  //F_CPU expected to be 16000000;
  }
  uint8_t clkpr = CLKPR;                                                //The clock prescaler register allows to divide the real clock to reduce power.
  if (clkpr & 0x80)                                                     //Clock prescaler enabled?
  {
    clkpr &= 0x0F;                                                      //Only the 4 lowest bits are valid.
    if (clkpr < 0x09) while (clkpr)                                     //Clock divider 0 ... 8 are valid.
    {
      ioClk >>= 1;                                                      //Divide IOclock by 2.
      --clkpr;
    }
  }
  return ioClk;
}







/********************************************************************************************************
 *    Calculate the timer Top32 divider value based on a wanted RPM and motor steps per revolution.
 *    [in]  rpm               Desired revolutions per minute.
 *          steps_revolution  Amount of steps required to rotate the shaft once. This is the amount of steps according to the motor datasheet
 *                            multiplied by the mode. Ex. a 200step/revolution motor in half step mode requires 400steps/revolution.
 *    [out] uint32_t          Virtual 32bit Timer TOP register; can be written to the stepMotorxTimerUpdate function.
 *    Note about the calculation:
 *    -   steps_per_minute = RPM * steps_per_revolution.
 *    -   steps_per_second = steps_per_minute / 60 * 2 (because there are 2 timer overflows required per step.
 *    -   to reduce the required calculation time, the source divides by 30 immideately, since AVR is slow using floats.
 *    -   timer = oscillator / steps_per_second, returned as a virtual 32bit value.
 */
uint32_t
motorTimerRPMtoTimerTop32 (float        rpm,
                           unsigned int steps_revolution)
{
  if (rpm <= 0.) return 0UL;                                            //0 Stops the motor.
  if (steps_revolution < 16) return 0UL;                                //Minimal supported steps per revolution.
  float steps_min = rpm * (float) (steps_revolution);                   //Steps per minute.
  float steps_per_s = steps_min / 30.;                                  //Steps per second (we count 30s because the timer must count 2x as fast).
  float oscillator = (float) (motorTimerGetIOclocksPerSecond ());       //System frequency.
  float timer = oscillator / steps_per_s;
  timer += .5;                                                          //Round the timer result in steps/s.
  return (uint32_t) (timer);                                            //Return as long unsigned integer.
}





/********************************************************************************************************
 *    - Initializes the motor driver step pulse pin as output. Required to output the step pulses to the A4988 driver.
 *    - Sets the A4988 micro step control, reset, sleep & enable pins as output.
 *    - Microstep Mode is set to FULL.
 *    - Microstep counter is set to the base counter value.
 *    [out] byte    B0000.0001 if Timer 1 initialized;
 *                  B0000.0010 if Timer 3 initialized;
 *                  B0000.0100 if Timer 4 initialized;
 *                  B0000.1000 if Timer 5 initialized (or any combination if multiple timers are used in parallel).
 */
int8_t
stepMotorInit ()
{
  int8_t status = 0;
  motorTimUpdateFlag = 0;                                               //No OCRxA update / no micro step update.
  #ifdef MOTOR_TIMER_1
    #ifdef A4988_1_SLEEP
      pinMode ( A4988_1_SLEEP, OUTPUT);
    #endif
    #ifdef A4988_1_MS1
      pinMode (   A4988_1_MS1, OUTPUT);
    #endif
    #ifdef A4988_1_MS2
      pinMode (   A4988_1_MS2, OUTPUT);
    #endif
    #ifdef A4988_1_MS3
      pinMode (   A4988_1_MS3, OUTPUT);
    #endif
    pinMode (   A4988_1_DIR, OUTPUT);
    #ifdef A4988_1_ENABLE
      pinMode (A4988_1_ENABLE, OUTPUT);
    #endif
    #ifdef A4988_1_RESET
      pinMode ( A4988_1_RESET, OUTPUT);
    #endif
    #if ((defined (__AVR_ATmega2560__) || defined (__AVR_ATmega2561__)))
      pinMode (11, OUTPUT);                                             //Mega.
    #else
      pinMode ( 9, OUTPUT);                                             //Uno.
    #endif
    motor1posCount = 0L;
    motor1micrStepMode = MOTORSTEP_FULL;
    stepMotor1SetStepMode (MOTORSTEP_FULL, MOTORUPDATE_NOW);
    motor1micrStepCount = motor1micrStepBaseCount;
    stepMotor1Reset ();
    status |= MOTOR1POSLO;
  #endif
  #if ((defined (__AVR_ATmega2560__) || defined (__AVR_ATmega2561__)))  //Mega only.
    #ifdef MOTOR_TIMER_3
      #ifdef A4988_3_SLEEP
        pinMode ( A4988_3_SLEEP, OUTPUT);
      #endif
      #ifdef A4988_3_MS1
        pinMode (   A4988_3_MS1, OUTPUT);
      #endif
      #ifdef A4988_3_MS2
        pinMode (   A4988_3_MS2, OUTPUT);
      #endif
      #ifdef A4988_3_MS3
        pinMode (   A4988_3_MS3, OUTPUT);
      #endif
      pinMode (   A4988_3_DIR, OUTPUT);
      #ifdef A4988_3_ENABLE
        pinMode (A4988_3_ENABLE, OUTPUT);
      #endif
      #ifdef A4988_3_RESET
        pinMode ( A4988_3_RESET, OUTPUT);
      #endif
      pinMode ( 5, OUTPUT);
      motor3posCount = 0L;
      motor3micrStepMode = MOTORSTEP_FULL;
      stepMotor3SetStepMode (MOTORSTEP_FULL, MOTORUPDATE_NOW);
      motor3micrStepCount = motor3micrStepBaseCount;
      stepMotor3Reset ();
      status |= MOTOR3POSLO;
    #endif
    #ifdef MOTOR_TIMER_4
      #ifdef A4988_4_SLEEP
        pinMode ( A4988_4_SLEEP, OUTPUT);
      #endif
      #ifdef A4988_4_MS1
        pinMode (   A4988_4_MS1, OUTPUT);
      #endif
      #ifdef A4988_4_MS2
        pinMode (   A4988_4_MS2, OUTPUT);
      #endif
      #ifdef A4988_4_MS3
        pinMode (   A4988_4_MS3, OUTPUT);
      #endif
      pinMode (   A4988_4_DIR, OUTPUT);
      #ifdef A4988_4_ENABLE
        pinMode (A4988_4_ENABLE, OUTPUT);
      #endif
      #ifdef A4988_4_RESET
        pinMode ( A4988_4_RESET, OUTPUT);
      #endif
      pinMode ( 6, OUTPUT);
      motor4posCount = 0L;
      motor4micrStepMode = MOTORSTEP_FULL;
      stepMotor4SetStepMode (MOTORSTEP_FULL, MOTORUPDATE_NOW);
      motor4micrStepCount = motor4micrStepBaseCount;
      stepMotor4Reset ();
      status |= MOTOR4POSLO;
    #endif
    #ifdef MOTOR_TIMER_5
      #ifdef A4988_5_SLEEP
        pinMode ( A4988_5_SLEEP, OUTPUT);
      #endif
      #ifdef A4988_5_MS1
        pinMode (   A4988_5_MS1, OUTPUT);
      #endif
      #ifdef A4988_5_MS2
        pinMode (   A4988_5_MS2, OUTPUT);
      #endif
      #ifdef A4988_5_MS3
        pinMode (   A4988_5_MS3, OUTPUT);
      #endif
      pinMode (   A4988_5_DIR, OUTPUT);
      #ifdef A4988_5_ENABLE
        pinMode (A4988_5_ENABLE, OUTPUT);
      #endif
      #ifdef A4988_5_RESET
        pinMode ( A4988_5_RESET, OUTPUT);
      #endif
      pinMode (46, OUTPUT);
      motor5posCount = 0L;
      motor5micrStepMode = MOTORSTEP_FULL;
      stepMotor5SetStepMode (MOTORSTEP_FULL, MOTORUPDATE_NOW);
      motor5micrStepCount = motor5micrStepBaseCount;
      stepMotor5Reset ();
      status |= MOTOR5POSLO;
    #endif
  #else                                                                 //UNO.
    #if (defined (MOTOR_TIMER_3) || defined (MOTOR_TIMER_4) || defined (MOTOR_DRIVER_5))
      status = -1;
    #endif
  #endif
  return status;
}





/********************************************************************************************************
 *    Wakes up the A4988 (in case it was sleeping) and sets the A4988 Enable pin (except if timTop32 == 0, or an error occured. In
 *    that case the A4988 config is not changed).
 *    Ex. Arduino Mega pin D5 drives an A4988 stepper motor pulses input. This pin is called on the Mega AVR
 *    PE3 and has as alternative timer function OC3A. Timer3 is used in Fast PWM mode. In this mode, only output compare register A
 *    for that timer (table 17-4) can be used to toggle OC3A pin (OC3B and OC3C are not used in Fast PMW).
 *    Every time the timer reaches this 16bit top value, OC3A pin is toggled (the actual 0 or 1 state inverts).
 *    Toggle has the advantage the resulting frequency has equal low/high time (called 50% duty cycle).
 *    Since there are 1 (Uno) and 4 (Mega) 16bit timers in the AVR and each time OCxA pin is available, this driver can control up
 *    to four A4988 drivers independent from each-other.
 *    [in] timTop32  A virtual 32bit IO clock divider required to obtain the amount of wanted motor steps per second.
 *                   Maximal value: (1024 * 65536) - 1 (otherwise error is returned).
 *                   Minimal value: 0 or >= 1600 (0 stops the motor pulses, 1600 applies 10kHz).
 *    [out] byte     -  1: timer running.
 *                   -  0: motor stopped.
 *                   - -1: timer value too large.
 *    Example: the user calls this function with parameter 4159231. The I/O clock is 16MHz.
 *    One would expected 16M / 4159231 = 3.846865 timer overruns per second.
 *    4159231 = 0x003F.76FF hexadecimal (the dot is only placed to separate the upper and lower 16bit values.
 *    The first check just tests if the value is less than 1024 * 65535, the maximal range of the timer with maximal prescaler.
 *     if ((timTop32 > MAXTIM32BIT) != 0) is executed as            0x003F.76FF
 *                                                       Compare >= 0xFC00.0000
 *                                                           Result             (smaller, so the 'if' is skipped, return not executed).
 *     if ((timTop32 & 0xFFFF0000UL) != 0)                          0x003F.76FF
 *                                                              AND 0xFFFF.0000
 *                                                           Result 0x003F.0000 (non zero, the if is executed).
 *       timTop32 >>= 3;                                            0x003F.76FF (gets divided by 8; becomes 0x0007.EEDF).
 *       cs = (1 << CS31);                                                      (Select Timer3 prescaler 8).
 *     if ((timTop32 & 0xFFFF0000UL) != 0)                          0x0007.EEDF
 *                                                              AND 0xFFFF.0000
 *                                                          Result: 0x0007.0000 (non zero, the if is executed).
 *       timTop32 >>= 3;                                            0x0007.EEDF (gets divided by 8; becomes 0x0000.FDD8).
 *       cs = (1 << CS31) | (1 << CS30)                                         (Select Timer3 prescaler 64).
 *     if ((timTop32 & 0xFFFF0000UL) != 0)                          0x0000.FDD8
 *                                                              AND 0xFFFF.0000
 *                                                          Result  0x0000.0000 (no more if's are executed).
 *    So the software selects prescaler 64 and writes the timer maximal value as 0xFDDB. This is 64987.
 *    The real timer top is 64 * 64987 = 4159168. That is 99.9985% accurate.
 *    Timer values below 0xFFFF will have zero deviation, since all fit in the 16bit Timer. Once the timer value exceeds 0xFFFF (=65535)
 *    some minor rounding errors may occur because the timer has a fixed prescaler of 1, 8, 64, 256 or 1024.
 *    Note the function exist for each timer. On a Uno, one can only use Timer1. On a Mega, one can control up to 4 motors independent
 *    from each-other (just enable one timer per motor).
 */
#ifdef MOTOR_TIMER_1
  int8_t
  stepMotor1TimerUpdate (uint32_t  timTop32)                            //Virtual 32-bit timer data TOP value.
  {
    if (timTop32 > TIM32BITMAX) return -1;                              //Doesn't fit in 10bit prescaler + 16bit timer (1024 * 65536).
    if (timTop32 == 0UL)
    {
      TCCR1B = (1 << WGM13) | (1 << WGM12);                             //Stop timer clock input.
      motorEndPos |= MOTOR1POSLO;                                       //Endpos reached.
      return 0;
    }
    motorEndPos &= ~MOTOR1POSLO;                                        //Endpos not reached.
    uint8_t sync = 1;
    if (TCCR1B & ((1 << CS12) | (1 << CS11) | (1 << CS10))) sync = 0;   //Timer off: no need to sync.
    if (timTop32 < 800UL) timTop32 = 1600UL;                            //Timer below 800 overflows/s not allowed.
    uint8_t cs = (1 << CS10);                                           //Divider 1.
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 3;                                                   //Divide by 8.
      cs = (1 << CS11);                                                 //Prescaler 8.
    }
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 3;                                                   //Divide previous result by 8 (div 64 total).
      cs = (1 << CS11) | (1 << CS10);                                   //Prescaler 64.
    }
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 2;                                                   //Divide previous result by 4 (64 * 4 = 256).
      cs = (1 << CS12);                                                 //Prescaler 256.
    }
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 2;                                                   //Divide previous result by 4 (256 * 4 = 1024).
      cs = (1 << CS12) | (1 << CS10);                                   //Prescaler 1024.
    }
    TCCR1A = (1 << COM1A0) | (1 << WGM11) | (1 << WGM10);               //COM1A0 selects 'toggle on compare match'. WGM lower 2 bits are b11.
    TCCR1B = (1 << WGM13) | (1 << WGM12);                               //Select already WGM=15.
    if ((timTop32 & 0xFFFF0000UL) != 0) return -1;                      //Error: 32bit timer cannot exceed 1024 (max prescaler) * 65536 (max 16bit).
    stepMotor1WakeUp ();
    stepMotor1DriverEnable ();
    if (timTop32 == 0UL) timTop32 = 0x0000FFFFUL;                       //Make sure the timer register is never zero.
    uint8_t lo = (uint8_t) (timTop32);                                  //Timer TOP value lowest byte.
    timTop32 >>= 8;                                                     //Shift the high byte of remaining 16bit to the lowest 8bit.
    uint8_t hi = (uint8_t) (timTop32);                                  //This is the upper byte of the 16bit hardware timer.
    if (sync != 0)
    {
      motor1ocah = hi;
      motor1ocal = lo;
      motorTimUpdateFlag |=  MOTOR1POSLO;                               //Set update request for the timer.
    }
    else
    {
      OCR1AH = hi;                                                      //Write the upper byte to the 16bit timer.
      OCR1AL = lo;                                                      //Write the low byte to the 16bit timer.
      TIMSK1 |= (1 << TOIE1);                                           //Enable Timer Overflow Interrupt Enable.
    }
    TCCR1B = (1 << WGM13) | (1 << WGM12) | cs;                          //Timer clock is started by writing prescaler clock divider selection.
    return 1;                                                           //Return "no error, motor spinning.
  }





  ISR (TIMER1_OVF_vect)                                                 //Timer/Counter1 Overflow.
  {
    --motor1micrStepCount;
    if (motor1micrStepCount == 0)                                       //Determine how many steps are required for 1 full step.
    {
      motor1micrStepCount = motor1micrStepBaseCount;
      if (motorDir & MOTOR1POSLO) --motor1posCount;
      else                        ++motor1posCount;
      if (motorTimUpdateFlag & MOTOR1POSLO)                             //Update request for the timer?
      {
        OCR1AH = motor1ocah;
        OCR1AL = motor1ocal;
        motorTimUpdateFlag &= ~MOTOR1POSLO;                            //Reset update request for the timer.
      }
      if (motorTimUpdateFlag & MOTOR1POSHI)
      {
        stepMotor1SetStepMode (motor1micrStepMode, MOTORUPDATE_NOW);
        motorTimUpdateFlag &= ~MOTOR1POSHI;
      }
      if (   (motorEndPos & MOTOR1POSHI)                                //End Position is set.
          && (motor1targetCount == motor1posCount))                     //End Position reached.
      {
        TIMSK1 &= ~(1 << TOIE1);                                        //Disalbe timer interrupt.
        TCCR1B &= ~(  (1 << CS12)                                       //Disable the timer.
                    | (1 << CS11)
                    | (1 << CS10));
        motorEndPos |= MOTOR1POSLO;                                     //Position reached.
      }
    }
  }
#endif





#ifdef MOTOR_TIMER_3
  int8_t
  stepMotor3TimerUpdate (uint32_t timTop32)                             //Virtual 32-bit timer data TOP value.
  {
    if (timTop32 > TIM32BITMAX) return -1;                              //Doesn't fit in 10bit prescaler + 16bit timer (1024 * 65536).
    if (timTop32 == 0UL)
    {
      TCCR3B = (1 << WGM33) | (1 << WGM32);                             //Stop timer clock input.
      motorEndPos |= MOTOR3POSLO;                                       //Endpos reached.
      return 0;
    }
    motorEndPos &= ~MOTOR3POSLO;                                        //Endpos not reached.
    uint8_t sync = 1;
    if (TCCR3B & ((1 << CS32) | (1 << CS31) | (1 << CS30))) sync = 0;   //Timer off: no need to sync.
    if (timTop32 < 800UL) timTop32 = 1600UL;                            //Timer below 800 overflows/s not allowed.
    uint8_t cs = (1 << CS30);                                           //Divider 1.
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 3;                                                   //Divide by 8.
      cs = (1 << CS31);                                                 //Prescaler 8.
    }
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 3;                                                   //Divide previous result by 8 (div 64 total).
      cs = (1 << CS31) | (1 << CS30);                                   //Prescaler 64.
    }
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 2;                                                   //Divide previous result by 4 (64 * 4 = 256).
      cs = (1 << CS32);                                                 //Prescaler 256.
    }
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 2;                                                   //Divide previous result by 4 (256 * 4 = 1024).
      cs = (1 << CS32) | (1 << CS30);                                   //Prescaler 1024.
    }
    TCCR3A = (1 << COM3A0) | (1 << WGM31) | (1 << WGM30);               //COM3A0 selects 'toggle on compare match'. WGM lower 2 bits are b11.
    if ((timTop32 & 0xFFFF0000UL) != 0) return -1;                      //Error: 32bit timer cannot exceed 1024 (max prescaler) * 65536 (max 16bit).
    stepMotor3WakeUp ();
    stepMotor3DriverEnable ();
    if (timTop32 == 0UL) timTop32 = 0x0000FFFFUL;                       //Make sure the timer register is never zero.
    TCCR3B = (1 << WGM33) | (1 << WGM32) | cs;                          //WGM upper 2 bits must be b11, combined is WGM = 15 (Fast PWM mode).
    uint8_t lo = (uint8_t) (timTop32);                                  //Timer TOP value lowest byte.
    timTop32 >>= 8;                                                     //Shift the high byte of remaining 16bit to the lowest 8bit.
    cs = (uint8_t) (timTop32);                                          //This is the upper byte of the 16bit hardware timer.
    if (sync != 0)
    {
      motor3ocah = cs;
      motor3ocal = lo;
      motorTimUpdateFlag |= MOTOR3POSLO;                                //Set update request for the timer.
    }
    else
    {
      OCR3AH = cs;                                                      //Write the upper byte to the 16bit timer.
      OCR3AL = lo;                                                      //Write the low byte to the 16bit timer.
      TIMSK3 |= (1 << TOIE3);                                           //Enable Timer Overflow Interrupt Enable.
    }
    return 1;                                                           //Return "no error, motor spinning.
  }





  ISR (TIMER3_OVF_vect)                                                 //Timer/Counter3 Overflow.
  {
    --motor3micrStepCount;
    if (motor3micrStepCount == 0)
    {
      motor3micrStepCount = motor3micrStepBaseCount;
      if (motorDir & MOTOR3POSLO) --motor3posCount;
      else                        ++motor3posCount;
      if (motorTimUpdateFlag & MOTOR3POSLO)                             //Update request for the timer?
      {
        OCR3AH = motor3ocah;
        OCR3AL = motor3ocal;
        motorTimUpdateFlag &= ~MOTOR3POSLO;                             //Reset update request for the timer.
      }
      if (motorTimUpdateFlag & MOTOR3POSHI)
      {
        stepMotor3SetStepMode (motor3micrStepMode, MOTORUPDATE_NOW);
        motorTimUpdateFlag &= ~MOTOR3POSHI;
      }
      if (   (motorEndPos & MOTOR3POSHI)                                //End Position is set.
          && (motor3targetCount == motor3posCount))                     //End Position reached.
      {
        TIMSK3 &= ~(1 << TOIE3);                                        //Disalbe timer interrupt.
        TCCR3B &= ~(  (1 << CS32)                                       //Disable the timer.
                    | (1 << CS31)
                    | (1 << CS30));
        motorEndPos |= MOTOR3POSLO;                                     //Position reached.
      }
    }
  }
#endif





#ifdef MOTOR_TIMER_4
  int8_t
  stepMotor4TimerUpdate (uint32_t timTop32)                             //Virtual 32-bit timer data TOP value.
  {
    if (timTop32 > TIM32BITMAX) return -1;                              //Doesn't fit in 10bit prescaler + 16bit timer (1024 * 65536).
    if (timTop32 == 0UL)
    {
      TCCR4B = (1 << WGM43) | (1 << WGM42);                             //Stop timer clock input.
      motorEndPos |= MOTOR4POSLO;                                       //Endpos reached.
      return 0;
    }
    motorEndPos &= ~MOTOR4POSLO;                                        //Endpos not reached.
    uint8_t sync = 1;
    if (TCCR4B & ((1 << CS42) | (1 << CS41) | (1 << CS40))) sync = 0;   //Timer off: no need to sync.
    if (timTop32 < 800UL) timTop32 = 1600UL;                            //Timer below 800 overflows/s not allowed.
    uint8_t cs = (1 << CS40);                                           //Divider 1.
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 3;                                                   //Divide by 8.
      cs = (1 << CS41);                                                 //Prescaler 8.
    }
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 3;                                                   //Divide previous result by 8 (div 64 total).
      cs = (1 << CS41) | (1 << CS40);                                   //Prescaler 64.
    }
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 2;                                                   //Divide previous result by 4 (64 * 4 = 256).
      cs = (1 << CS42);                                                 //Prescaler 256.
    }
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 2;                                                   //Divide previous result by 4 (256 * 4 = 1024).
      cs = (1 << CS42) | (1 << CS40);                                   //Prescaler 1024.
    }
    TCCR4A = (1 << COM4A0) | (1 << WGM41) | (1 << WGM40);               //COM4A0 selects 'toggle on compare match'. WGM lower 2 bits are b11.
    if ((timTop32 & 0xFFFF0000UL) != 0) return -1;                      //Error: 32bit timer cannot exceed 1024 (max prescaler) * 65536 (max 16bit).
    stepMotor4WakeUp ();
    stepMotor4DriverEnable ();
    if (timTop32 == 0UL) timTop32 = 0x0000FFFFUL;                       //Make sure the timer register is never zero.
    TCCR4B = (1 << WGM43) | (1 << WGM42) | cs;                          //WGM upper 2 bits must be b11, combined is WGM = 15 (Fast PWM mode).
    uint8_t lo = (uint8_t) (timTop32);                                  //Timer TOP value lowest byte.
    timTop32 >>= 8;                                                     //Shift the high byte of remaining 16bit to the lowest 8bit.
    cs = (uint8_t) (timTop32);                                          //This is the upper byte of the 16bit hardware timer.
    if (sync != 0)
    {
      motor4ocah = cs;
      motor4ocal = lo;
      motorTimUpdateFlag |= MOTOR4POSLO;                                //Set update request for the timer.
    }
    else
    {
      OCR4AH = cs;                                                      //Write the upper byte to the 16bit timer.
      OCR4AL = lo;                                                      //Write the low byte to the 16bit timer.
      TIMSK4 |= (1 << TOIE4);                                           //Enable Timer Overflow Interrupt Enable.
    }
    return 1;                                                           //Return "no error, motor spinning.
  }





  ISR (TIMER4_OVF_vect)                                                 //Timer/Counter4 Overflow.
  {
    --motor4micrStepCount;
    if (motor4micrStepCount == 0)
    {
      motor4micrStepCount = motor4micrStepBaseCount;
      if (motorDir & MOTOR4POSLO) --motor4posCount;
      else                        ++motor4posCount;
      if (motorTimUpdateFlag & MOTOR4POSLO)                             //Update request for the timer?
      {
        OCR4AH = motor4ocah;
        OCR4AL = motor4ocal;
        motorTimUpdateFlag &= ~MOTOR4POSLO;                             //Reset update request for the timer.
      }
      if (motorTimUpdateFlag & MOTOR4POSHI)
      {
        stepMotor4SetStepMode (motor4micrStepMode, MOTORUPDATE_NOW);
        motorTimUpdateFlag &= ~MOTOR4POSHI;
      }
      if (   (motorEndPos & MOTOR4POSHI)                                //End Position is set.
          && (motor4targetCount == motor4posCount))                     //End Position reached.
      {
        TIMSK4 &= ~(1 << TOIE4);                                        //Disalbe timer interrupt.
        TCCR4B &= ~(  (1 << CS42)                                       //Disable the timer.
                    | (1 << CS41)
                    | (1 << CS40));
        motorEndPos |= MOTOR4POSLO;                                     //Position reached.
      }
    }
  }
#endif





#ifdef MOTOR_TIMER_5
  int8_t
  stepMotor5TimerUpdate (uint32_t timTop32)                             //Virtual 32-bit timer data TOP value.
  {
    if (timTop32 > TIM32BITMAX) return -1;                              //Doesn't fit in 10bit prescaler + 16bit timer (1024 * 65536).
    if (timTop32 == 0UL)
    {
      TCCR5B = (1 << WGM53) | (1 << WGM52);                             //Stop timer clock input.
      motorEndPos |= MOTOR5POSLO;                                       //Endpos reached.
      return 0;
    }
    motorEndPos &= ~MOTOR5POSLO;                                        //Endpos not reached.
    uint8_t sync = 1;
    if (TCCR5B & ((1 << CS52) | (1 << CS51) | (1 << CS50))) sync = 0;   //Timer off: no need to sync.
    if (timTop32 < 800UL) timTop32 = 1600UL;                            //Timer below 800 overflows/s not allowed.
    uint8_t cs = (1 << CS50);                                           //Divider 1.
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 3;                                                   //Divide by 8.
      cs = (1 << CS51);                                                 //Prescaler 8.
    }
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 3;                                                   //Divide previous result by 8 (div 64 total).
      cs = (1 << CS51) | (1 << CS50);                                   //Prescaler 64.
    }
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 2;                                                   //Divide previous result by 4 (64 * 4 = 256).
      cs = (1 << CS52);                                                 //Prescaler 256.
    }
    if ((timTop32 & 0xFFFF0000UL) != 0)                                 //Doesn't fit in 16-bit.
    {
      timTop32 >>= 2;                                                   //Divide previous result by 4 (256 * 4 = 1024).
      cs = (1 << CS52) | (1 << CS50);                                   //Prescaler 1024.
    }
    TCCR5A = (1 << COM5A0) | (1 << WGM51) | (1 << WGM50);               //COM3A0 selects 'toggle on compare match'. WGM lower 2 bits are b11.
    if ((timTop32 & 0xFFFF0000UL) != 0) return -1;                      //Error: 32bit timer cannot exceed 1024 (max prescaler) * 65536 (max 16bit).
    stepMotor5WakeUp ();
    stepMotor5DriverEnable ();
    if (timTop32 == 0UL) timTop32 = 0x0000FFFFUL;                       //Make sure the timer register is never zero.
    TCCR5B = (1 << WGM53) | (1 << WGM52) | cs;                          //WGM upper 2 bits must be b11, combined is WGM = 15 (Fast PWM mode).
    uint8_t lo = (uint8_t) (timTop32);                                  //Timer TOP value lowest byte.
    timTop32 >>= 8;                                                     //Shift the high byte of remaining 16bit to the lowest 8bit.
    cs = (uint8_t) (timTop32);                                          //This is the upper byte of the 16bit hardware timer.
    if (sync != 0)
    {
      motor5ocah = cs;
      motor5ocal = lo;
      motorTimUpdateFlag |= MOTOR5POSLO;                                //Set update request for the timer.
    }
    else
    {
      OCR5AH = cs;                                                      //Write the upper byte to the 16bit timer.
      OCR5AL = lo;                                                      //Write the low byte to the 16bit timer.
      TIMSK5 |= (1 << TOIE5);                                           //Enable Timer Overflow Interrupt Enable.
    }
    return 1;                                                           //Return "no error, motor spinning.
  }





  ISR (TIMER5_OVF_vect)                                                 //Timer/Counter5 Overflow.
  {
    --motor5micrStepCount;
    if (motor5micrStepCount == 0)
    {
      motor5micrStepCount = motor5micrStepBaseCount;
      if (motorDir & MOTOR5POSLO) --motor5posCount;
      else                        ++motor5posCount;
      if ((motorTimUpdateFlag & MOTOR5POSLO) != 0)                      //Timer OCRxAH/L Update request?
      {
        OCR5AH = motor5ocah;
        OCR5AL = motor5ocal;
        motorTimUpdateFlag &= ~MOTOR5POSLO;                             //Clear the request.
      }
      if (motorTimUpdateFlag & MOTOR5POSHI)
      {
        stepMotor5SetStepMode (motor5micrStepMode, MOTORUPDATE_NOW);
        motorTimUpdateFlag &= ~MOTOR5POSHI;
      }
      if (   ((motorEndPos & MOTOR5POSHI) != 0)                         //End Position set?
          && (motor5targetCount == motor5posCount))                     //End Position reached.
      {
        TIMSK5 &= ~(1 << TOIE5);                                        //Disalbe timer interrupt.
        TCCR5B &= ~(  (1 << CS52)                                       //Disable the timer.
                    | (1 << CS51)
                    | (1 << CS50));
        motorEndPos |= MOTOR5POSLO;                                     //Flag Position reached.
      }
    }
  }
#endif






/********************************************************************************************************
 *    Set the step mode for any of the motor(s). The default function excection will remember the settling but not apply it immediately
 *    in hardware. Reason is that one should change the A4988 microstep mode only at a moment where present mode and new selected mode
 *    allow the active driven output state. Therefor, microsteps are only changed after a full step move has been detected.
 *    [in] sel   - MOTORSTEP_FULL        2 Phase.
 *               - MOTORSTEP_HALF        1-2 Phase
 *               - MOTORSTEP_QUARTER     W1-2 Phase.
 *               - MOTORSTEP_EIGHT       2W1-2 Phase.
 *               - MOTORSTEP_SIXTEENTH   4W1-2 Phase.
 *          u    - MOTORUPDATE_SYNCHRONIZED waits for the 1st full step to be finished before changing the MS pins of the A4988.
 *               - MOTORUPDATE_NOW          only to be used during init or IRQ service: writes the micro stepping pins now.
 *    [out] - 0  Motor Microstepping set without delay.
 *          - 1  Microstepping pins will be set on the next equivalent of a full step.
 */
#ifdef MOTOR_TIMER_1
  int8_t
  stepMotor1SetStepMode (MOTORSTEPSELECTION sel,
                         MOTORUPDATE u)
  {
    if (u == MOTORUPDATE_SYNCHRONIZED)                                  //Synchronized waits for full step finished before changing mode.
    {
      motor1micrStepMode = sel;
      motorTimUpdateFlag |= MOTOR1POSHI;
      return 1;
    }
    switch (sel)
    {
    case MOTORSTEP_FULL:        {
                                  #ifdef A4988_1_MS1
                                    digitalWrite (A4988_1_MS1, LOW);
                                  #endif
                                  #ifdef A4988_1_MS2
                                    digitalWrite (A4988_1_MS2, LOW);
                                  #endif
                                  #ifdef A4988_1_MS3
                                    digitalWrite (A4988_1_MS3, LOW);
                                  #endif
                                }
                                motor1micrStepBaseCount = 2;            //2 IRQs per A4988 step.
                                break;
    case MOTORSTEP_HALF:        {
                                  #ifdef A4988_1_MS1
                                    digitalWrite (A4988_1_MS1, HIGH);
                                  #endif
                                  #ifdef A4988_1_MS2
                                    digitalWrite (A4988_1_MS2, LOW);
                                  #endif
                                  #ifdef A4988_1_MS3
                                    digitalWrite (A4988_1_MS3, LOW);
                                  #endif
                                }
                                motor1micrStepBaseCount = 4;
                                break;
    case MOTORSTEP_QUARTER:     {
                                  #ifdef A4988_1_MS1
                                    digitalWrite (A4988_1_MS1, LOW);
                                  #endif
                                  #ifdef A4988_1_MS2
                                    digitalWrite (A4988_1_MS2, HIGH);
                                  #endif
                                  #ifdef A4988_1_MS3
                                    digitalWrite (A4988_1_MS3, LOW);
                                  #endif
                                }
                                motor1micrStepBaseCount = 8;
                                break;
    case MOTORSTEP_EIGHT:       {
                                  #ifdef A4988_1_MS1
                                    digitalWrite (A4988_1_MS1, HIGH);
                                  #endif
                                  #ifdef A4988_1_MS2
                                    digitalWrite (A4988_1_MS2, HIGH);
                                  #endif
                                  #ifdef A4988_1_MS3
                                    digitalWrite (A4988_1_MS3, LOW);
                                  #endif
                                }
                                motor1micrStepBaseCount = 16;
                                break;
    case MOTORSTEP_SIXTEENTH:   {
                                  #ifdef A4988_1_MS1
                                    digitalWrite (A4988_1_MS1, LOW);
                                  #endif
                                  #ifdef A4988_1_MS2
                                    digitalWrite (A4988_1_MS2, LOW);
                                  #endif
                                  #ifdef A4988_1_MS3
                                    digitalWrite (A4988_1_MS3, HIGH);
                                  #endif
                                }
                                motor1micrStepBaseCount = 32;
                                break;
    default:                    return -1;
                                break;
    }
    return 0;
  }
#endif

#ifdef MOTOR_TIMER_3
  int8_t
  stepMotor3SetStepMode (MOTORSTEPSELECTION sel,
                         MOTORUPDATE u)
  {
    if (u == MOTORUPDATE_SYNCHRONIZED)                                  //Synchronized waits for full step finished before changing mode.
    {
      motor3micrStepMode = sel;
      motorTimUpdateFlag |= MOTOR3POSHI;
      return 1;
    }
    switch (sel)
    {
    case MOTORSTEP_FULL:        {
                                  #ifdef A4988_3_MS1
                                    digitalWrite (A4988_3_MS1, LOW);
                                  #endif
                                  #ifdef A4988_3_MS2
                                    digitalWrite (A4988_3_MS2, LOW);
                                  #endif
                                  #ifdef A4988_3_MS3
                                    digitalWrite (A4988_3_MS3, LOW);
                                  #endif
                                }
                                motor3micrStepBaseCount = 2;            //2 IRQs per A4988 step.
                                break;
    case MOTORSTEP_HALF:        {
                                  #ifdef A4988_3_MS1
                                    digitalWrite (A4988_3_MS1, HIGH);
                                  #endif
                                  #ifdef A4988_3_MS2
                                    digitalWrite (A4988_3_MS2, LOW);
                                  #endif
                                  #ifdef A4988_3_MS3
                                    digitalWrite (A4988_3_MS3, LOW);
                                  #endif
                                }
                                motor3micrStepBaseCount = 4;
                                break;
    case MOTORSTEP_QUARTER:     {
                                  #ifdef A4988_3_MS1
                                    digitalWrite (A4988_3_MS1, LOW);
                                  #endif
                                  #ifdef A4988_3_MS2
                                    digitalWrite (A4988_3_MS2, HIGH);
                                  #endif
                                  #ifdef A4988_3_MS3
                                    digitalWrite (A4988_3_MS3, LOW);
                                  #endif
                                }
                                motor3micrStepBaseCount = 8;
                                break;
    case MOTORSTEP_EIGHT:       {
                                  #ifdef A4988_3_MS1
                                    digitalWrite (A4988_3_MS1, HIGH);
                                  #endif
                                  #ifdef A4988_3_MS2
                                    digitalWrite (A4988_3_MS2, HIGH);
                                  #endif
                                  #ifdef A4988_3_MS3
                                    digitalWrite (A4988_3_MS3, LOW);
                                  #endif
                                }
                                motor3micrStepBaseCount = 16;
                                break;
    case MOTORSTEP_SIXTEENTH:   {
                                  #ifdef A4988_3_MS1
                                    digitalWrite (A4988_3_MS1, LOW);
                                  #endif
                                  #ifdef A4988_3_MS2
                                    digitalWrite (A4988_3_MS2, LOW);
                                  #endif
                                  #ifdef A4988_3_MS3
                                    digitalWrite (A4988_3_MS3, HIGH);
                                  #endif
                                }
                                motor3micrStepBaseCount = 32;
                                break;
    default:                    return -1;
                                break;
    }
    return 0;
  }
#endif

#ifdef MOTOR_TIMER_4
  int8_t
  stepMotor4SetStepMode (MOTORSTEPSELECTION sel,
                         MOTORUPDATE u)
  {
    if (u == MOTORUPDATE_SYNCHRONIZED)                                  //Synchronized waits for full step finished before changing mode.
    {
      motor4micrStepMode = sel;
      motorTimUpdateFlag |= MOTOR4POSHI;
      return 1;
    }
    switch (sel)
    {
    case MOTORSTEP_FULL:        {
                                  #ifdef A4988_4_MS1
                                    digitalWrite (A4988_4_MS1, LOW);
                                  #endif
                                  #ifdef A4988_4_MS2
                                    digitalWrite (A4988_4_MS2, LOW);
                                  #endif
                                  #ifdef A4988_4_MS3
                                    digitalWrite (A4988_4_MS3, LOW);
                                  #endif
                                }
                                motor4micrStepBaseCount = 2;            //2 IRQs per A4988 step.
                                break;
    case MOTORSTEP_HALF:        {
                                  #ifdef A4988_4_MS1
                                    digitalWrite (A4988_4_MS1, HIGH);
                                  #endif
                                  #ifdef A4988_4_MS2
                                    digitalWrite (A4988_4_MS2, LOW);
                                  #endif
                                  #ifdef A4988_4_MS3
                                    digitalWrite (A4988_4_MS3, LOW);
                                  #endif
                                }
                                motor4micrStepBaseCount = 4;
                                break;
    case MOTORSTEP_QUARTER:     {
                                  #ifdef A4988_4_MS1
                                    digitalWrite (A4988_4_MS1, LOW);
                                  #endif
                                  #ifdef A4988_4_MS2
                                    digitalWrite (A4988_4_MS2, HIGH);
                                  #endif
                                  #ifdef A4988_4_MS3
                                    digitalWrite (A4988_4_MS3, LOW);
                                  #endif
                                }
                                motor4micrStepBaseCount = 8;
                                break;
    case MOTORSTEP_EIGHT:       {
                                  #ifdef A4988_4_MS1
                                    digitalWrite (A4988_4_MS1, HIGH);
                                  #endif
                                  #ifdef A4988_4_MS2
                                    digitalWrite (A4988_4_MS2, HIGH);
                                  #endif
                                  #ifdef A4988_4_MS3
                                    digitalWrite (A4988_4_MS3, LOW);
                                  #endif
                                }
                                motor4micrStepBaseCount = 16;
                                break;
    case MOTORSTEP_SIXTEENTH:   {
                                  #ifdef A4988_4_MS1
                                    digitalWrite (A4988_4_MS1, LOW);
                                  #endif
                                  #ifdef A4988_4_MS2
                                    digitalWrite (A4988_4_MS2, LOW);
                                  #endif
                                  #ifdef A4988_4_MS3
                                    digitalWrite (A4988_4_MS3, HIGH);
                                  #endif
                                }
                                motor4micrStepBaseCount = 32;
                                break;
    default:                    return -1;
                                break;
    }
    return 0;
  }
#endif

#ifdef MOTOR_TIMER_5
  int8_t
  stepMotor5SetStepMode (MOTORSTEPSELECTION sel,
                         MOTORUPDATE u)
  {
    if (u == MOTORUPDATE_SYNCHRONIZED)                                  //Synchronized waits for full step finished before changing mode.
    {
      motor5micrStepMode = sel;
      motorTimUpdateFlag |= MOTOR5POSHI;
      return 1;
    }
    switch (sel)
    {
    case MOTORSTEP_FULL:        {
                                  #ifdef A4988_5_MS1
                                    digitalWrite (A4988_5_MS1, LOW);
                                  #endif
                                  #ifdef A4988_5_MS2
                                    digitalWrite (A4988_5_MS2, LOW);
                                  #endif
                                  #ifdef A4988_5_MS3
                                    digitalWrite (A4988_5_MS3, LOW);
                                  #endif
                                }
                                motor5micrStepBaseCount = 2;            //2 IRQs for each A4988 Full Step.
                                break;
    case MOTORSTEP_HALF:        {
                                  #ifdef A4988_5_MS1
                                    digitalWrite (A4988_5_MS1, HIGH);
                                  #endif
                                  #ifdef A4988_5_MS2
                                    digitalWrite (A4988_5_MS2, LOW);
                                  #endif
                                  #ifdef A4988_5_MS3
                                    digitalWrite (A4988_5_MS3, LOW);
                                  #endif
                                }
                                motor5micrStepBaseCount = 4;            //4 IRQs for each A4988 Full Step.
                                break;
    case MOTORSTEP_QUARTER:     {
                                  #ifdef A4988_5_MS1
                                    digitalWrite (A4988_5_MS1, LOW);
                                  #endif
                                  #ifdef A4988_5_MS2
                                    digitalWrite (A4988_5_MS2, HIGH);
                                  #endif
                                  #ifdef A4988_5_MS3
                                    digitalWrite (A4988_5_MS3, LOW);
                                  #endif
                                }
                                motor5micrStepBaseCount = 8;            //8 IRQs for each A4988 Full Step.
                                break;
    case MOTORSTEP_EIGHT:       {
                                  #ifdef A4988_5_MS1
                                    digitalWrite (A4988_5_MS1, HIGH);
                                  #endif
                                  #ifdef A4988_5_MS2
                                    digitalWrite (A4988_5_MS2, HIGH);
                                  #endif
                                  #ifdef A4988_5_MS3
                                    digitalWrite (A4988_5_MS3, LOW);
                                  #endif
                                  motor5micrStepBaseCount = 16;         //16 IRQs for each A4988 Full Step.
                                }
                                break;
    case MOTORSTEP_SIXTEENTH:   {
                                  #ifdef A4988_5_MS1
                                    digitalWrite (A4988_5_MS1, LOW);
                                  #endif
                                  #ifdef A4988_5_MS2
                                    digitalWrite (A4988_5_MS2, LOW);
                                  #endif
                                  #ifdef A4988_5_MS3
                                    digitalWrite (A4988_5_MS3, HIGH);
                                  #endif
                                }
                                motor5micrStepBaseCount = 32;           //32 IRQs for each A4988 Full Step.
                                break;
    default:                    return -1;
                                break;
    }
    return 0;
  }
#endif






/********************************************************************************************************
 *    Put the motor controller into sleep or wake it up.
 */
#ifdef MOTOR_TIMER_1
  void
  stepMotor1Sleep ()
  {
    #ifdef A4988_1_SLEEP
      digitalWrite (A4988_1_SLEEP, LOW);
    #endif
  }
  void
  stepMotor1WakeUp ()
  {
    #ifdef A4988_1_SLEEP
      if (outputPinReadDigital (A4988_1_SLEEP) == LOW)
      {
        digitalWrite (A4988_1_SLEEP, HIGH);
        delay (1);
      }
    #endif
  }
#endif

#ifdef MOTOR_TIMER_3
  void
  stepMotor3Sleep ()
  {
    #ifdef A4988_3_SLEEP
      digitalWrite (A4988_3_SLEEP, LOW);
    #endif
  }
  void
  stepMotor3WakeUp ()
  {
    #ifdef A4988_3_SLEEP
      if (outputPinReadDigital (A4988_3_SLEEP) == LOW)
      {
        digitalWrite (A4988_3_SLEEP, HIGH);
        delay (1);
      }
    #endif
  }
#endif

#ifdef MOTOR_TIMER_4
  void
  stepMotor4Sleep ()
  {
    #ifdef A4988_4_SLEEP
      digitalWrite (A4988_4_SLEEP, LOW);
    #endif
  }
  void
  stepMotor4WakeUp ()
  {
    #ifdef A4988_4_SLEEP
      if (outputPinReadDigital (A4988_4_SLEEP) == LOW)
      {
        digitalWrite (A4988_4_SLEEP, HIGH);
        delay (1);
      }
    #endif
  }
#endif

#ifdef MOTOR_TIMER_5
  void
  stepMotor5Sleep ()
  {
    #ifdef A4988_5_SLEEP
      digitalWrite (A4988_5_SLEEP, LOW);
    #endif
  }
  void
  stepMotor5WakeUp ()
  {
    #ifdef A4988_5_SLEEP
      if (outputPinReadDigital (A4988_5_SLEEP) == LOW)
      {
        digitalWrite (A4988_5_SLEEP, HIGH);
        delay (1);
      }
    #endif
  }
#endif





/********************************************************************************************************
 *    Enable or Disable the power driver stage of the A4988.
 */
#ifdef MOTOR_TIMER_1
  void
  stepMotor1DriverEnable ()
  {
    #ifdef A4988_1_ENABLE
      digitalWrite (A4988_1_ENABLE, HIGH);
    #endif
  }
  void
  stepMotor1DriverDisable ()
  {
    #ifdef A4988_1_ENABLE
      digitalWrite (A4988_1_ENABLE, LOW);
    #endif
  }
#endif
#ifdef MOTOR_TIMER_3
  void
  stepMotor3DriverEnable ()
  {
    #ifdef A4988_3_ENABLE
      digitalWrite (A4988_3_ENABLE, HIGH);
    #endif
  }
  void
  stepMotor3DriverDisable ()
  {
    #ifdef A4988_3_ENABLE
      digitalWrite (A4988_3_ENABLE, LOW);
    #endif
  }
#endif
#ifdef MOTOR_TIMER_4
  void
  stepMotor4DriverEnable ()
  {
    #ifdef A4988_4_ENABLE
      digitalWrite (A4988_4_ENABLE, HIGH);
    #endif
  }
  void
  stepMotor4DriverDisable ()
  {
    #ifdef A4988_4_ENABLE
      digitalWrite (A4988_4_ENABLE, LOW);
    #endif
  }
#endif
#ifdef MOTOR_TIMER_5
  void
  stepMotor5DriverEnable ()
  {
    #ifdef A4988_5_ENABLE
      digitalWrite (A4988_5_ENABLE, HIGH);
    #endif
  }
  void
  stepMotor5DriverDisable ()
  {
    #ifdef A4988_5_ENABLE
      digitalWrite (A4988_5_ENABLE, LOW);
    #endif
  }
#endif





/********************************************************************************************************
 *    Send a reset pulse to the motor driver.
 *    Set the timer clock prescaler to B000, to disables the Timer (no motor step pulses will be applied).
 *    Disable timer interrupts.
 *    Reset the step position counter.
 *    Disable end position detection.
 */
#ifdef MOTOR_TIMER_1
  void
  stepMotor1Reset ()
  {
    TIMSK1 &= ~(1 << TOIE1);
    TCCR1B &= ~(  (1 << CS12)
                | (1 << CS11)
                | (1 << CS10));
    #ifdef A4988_4_RESET
      digitalWrite (A4988_1_RESET, LOW);
      delay (1);
      digitalWrite (A4988_1_RESET, HIGH);
      delay (1);
    #endif
    motor1posCount = 0L;
    stepMotor1EndPointDisable ();
  }
#endif

#ifdef MOTOR_TIMER_3
  void
  stepMotor3Reset ()
  {
    TIMSK3 &= ~(1 << TOIE3);
    TCCR3B &= ~(  (1 << CS32)
                | (1 << CS31)
                | (1 << CS30));
    #ifdef A4988_4_RESET
      digitalWrite (A4988_3_RESET, LOW);
      delay (1);
      digitalWrite (A4988_3_RESET, HIGH);
      delay (1);
    #endif
    motor3posCount = 0L;
    stepMotor3EndPointDisable ();
  }
#endif

#ifdef MOTOR_TIMER_4
  void
  stepMotor4Reset ()
  {
    TIMSK4 &= ~(1 << TOIE4);
    TCCR4B &= ~(  (1 << CS42)
                | (1 << CS41)
                | (1 << CS40));
    #ifdef A4988_4_RESET
      digitalWrite (A4988_4_RESET, LOW);
      delay (1);
      digitalWrite (A4988_4_RESET, HIGH);
      delay (1);
    #endif
    motor4posCount = 0L;
    stepMotor4EndPointDisable ();
  }
#endif

#ifdef MOTOR_TIMER_5
  void
  stepMotor5Reset ()
  {
    TIMSK5 &= ~(1 << TOIE5);
    TCCR5B &= ~(  (1 << CS52)
                | (1 << CS51)
                | (1 << CS50));

    #ifdef A4988_5_RESET
      digitalWrite (A4988_5_RESET, LOW);
      delay (1);
      digitalWrite (A4988_5_RESET, HIGH);
      delay (1);
    #endif
    motor5posCount = 0L;
    stepMotor5EndPointDisable ();
  }
#endif





/********************************************************************************************************
 *    Set the direction pin for the active motor.
 *    [in] byte   0: direction pin is logic '0' = low.
 *               !0: direction pin is logic '1' = high.
 */
#ifdef MOTOR_TIMER_1
  void
  stepMotor1Direction (uint8_t dl)
  {
    digitalWrite (A4988_1_DIR, dl ? HIGH : LOW);
    if (dl) motorDir |= MOTOR1POSLO;
    else    motorDir &= ~MOTOR1POSLO;
  }
#endif

#ifdef MOTOR_TIMER_3
  void
  stepMotor3Direction (uint8_t dl)
  {
    digitalWrite (A4988_3_DIR, dl ? HIGH : LOW);
    if (dl) motorDir |= MOTOR3POSLO;
    else    motorDir &= ~MOTOR3POSLO;
  }
#endif

#ifdef MOTOR_TIMER_4
  void
  stepMotor4Direction (uint8_t dl)
  {
    digitalWrite (A4988_4_DIR, dl ? HIGH : LOW);
    if (dl) motorDir |= MOTOR4POSLO;
    else    motorDir &= ~MOTOR4POSLO;
  }
#endif

#ifdef MOTOR_TIMER_5
  void
  stepMotor5Direction (uint8_t dl)
  {
    digitalWrite (A4988_5_DIR, dl ? HIGH : LOW);
    if (dl) motorDir |= MOTOR5POSLO;
    else    motorDir &= ~MOTOR5POSLO;
  }
#endif






/********************************************************************************************************
 *    Enable or disable end point settling. This remains effective regardless how often the motor step rate or direction is
 *    changed. One needs to call the disable function to avoid the motor keeps on being stopped at the programmed position.
 *    Note: for every step where the motor direction is 0 and the equivalent of a full step, a 32bit position counter is incremented.
 *          Similarly, when direction is 1, for every equivalent of a full step the 32bit position counter is decremented.
 *          When this position counter matches endpoint, motor STEP pulses are stopped. The user can than call a new timer settling
 *          to start the motor again. Note the function does no error checking:
 *          - If one puts the maximal value (0x7FFFFFFF) on a 200steps/rotation motor, the end point is 10,737,418 revolutions from reset point.
 *          - If DIR = 0, the present position count is 1000 steps and the endpoint is set to 500 steps, the motor keeps virtually spinning forever.
 *            In reality, after the ~4000Million steps, a 32bit signed overflow has occured what will eventually stop the motor.
 *            -> The user has to make sure the endpoint is reached in reasonable time / is set correctly vs. the chosen direction and present position.
 *    [in]  endpoint: a 32bit signed value (positive: endpoint is XX steps from reset position with A4988 DIR = 0)
 *                                         (negative: endpoint is XX steps from reset position with A4988 DIR = 1)
 */
#ifdef MOTOR_TIMER_1
  void
  stepMotor1EndPointEnable (int32_t endpoint)
  {
    motor1targetCount = endpoint;
    motorEndPos |= MOTOR1POSHI;
  }
  void
  stepMotor1EndPointDisable ()
  {
    motorEndPos &= ~(  MOTOR1POSHI
                     | MOTOR1POSLO);
  }
#endif

#ifdef MOTOR_TIMER_3
  void
  stepMotor3EndPointEnable (int32_t endpoint)
  {
    motor3targetCount = endpoint;
    motorEndPos |= MOTOR3POSHI;
  }
  void
  stepMotor3EndPointDisable ()
  {
    motorEndPos &= ~(  MOTOR3POSHI
                     | MOTOR3POSLO);
  }
#endif

#ifdef MOTOR_TIMER_4
  void
  stepMotor4EndPointEnable (int32_t endpoint)
  {
    motor4targetCount = endpoint;
    motorEndPos |= MOTOR4POSHI;
  }
  void
  stepMotor4EndPointDisable ()
  {
    motorEndPos &= ~(  MOTOR4POSHI
                     | MOTOR4POSLO);
  }
#endif

#ifdef MOTOR_TIMER_5
  void
  stepMotor5EndPointEnable (int32_t endpoint)
  {
    motor5targetCount = endpoint;
    motorEndPos |= MOTOR5POSHI;
  }
  void
  stepMotor5EndPointDisable ()
  {
    motorEndPos &= ~(  MOTOR5POSHI
                     | MOTOR5POSLO);
  }
#endif





/********************************************************************************************************
 *    Return the present position count of the motor in full steps, discarding eventual microsteps.
 *    If the motor is operated in another step mode, this function returns the amount of full steps up to the moment of calling
 *    this function.
 *    [out] int32_t     Amount of full steps away from zero since the last A4988 reset.
 *                      - Positive indicates DIR was 0; negative indicates DIR was 1.
 */
#ifdef MOTOR_TIMER_1
  int32_t
  stepMotor1GetPosition ()
  {
    return motor1posCount;
  }
#endif

#ifdef MOTOR_TIMER_3
  int32_t
  stepMotor3GetPosition ()
  {
    return motor3posCount;
  }
#endif

#ifdef MOTOR_TIMER_4
  int32_t
  stepMotor4GetPosition ()
  {
    return motor4posCount;
  }
#endif

#ifdef MOTOR_TIMER_5
  int32_t
  stepMotor5GetPosition ()
  {
    return motor5posCount;
  }
#endif





/********************************************************************************************************
 *    Return the motor status:
 *    [Out] int8_t    10: A4988 active, motor spinnig,
 *                     1: A4988 active, motor stopped because of end point.
 *                     0: A4988 active, motor stopped,
 *                    -1: A4988 driver not enabled, motor stopped,
 *                    -2: A4988 driver sleeping.
 */
#ifdef MOTOR_TIMER_1
  int8_t
  stepMotor1Status ()
  {
    #ifdef A4988_1_ENABLE
      if (outputPinReadDigital (A4988_1_ENABLE) == LOW) return -1;
    #endif
    #ifdef A4988_1_SLEEP
      if (outputPinReadDigital (A4988_1_SLEEP) == LOW) return -2;
    #endif
    if (   (TIMSK1 & (1 << TOIE1))                                      //Timer interrupt enabled.
        && (TCCR1B & (  (1 << CS12)                                     //Timer has active clock.
                      | (1 << CS11)
                      | (1 << CS10))))
    {
      if (motorEndPos & MOTOR1POSLO) return 1;                          //Position reached.
      return 0;
    }
    return 10;
  }
#endif

#ifdef MOTOR_TIMER_3
  int8_t
  stepMotor3Status ()
  {
    #ifdef A4988_3_ENABLE
      if (outputPinReadDigital (A4988_3_ENABLE) == LOW) return -1;
    #endif
    #ifdef A4988_3_SLEEP
      if (outputPinReadDigital (A4988_3_SLEEP) == LOW) return -2;
    #endif
    if (   (TIMSK3 & (1 << TOIE3))                                      //Timer interrupt enabled.
        && (TCCR3B & (  (1 << CS32)                                     //Timer has active clock.
                      | (1 << CS31)
                      | (1 << CS30))))
    {
      if (motorEndPos & MOTOR3POSLO) return 1;                          //Position reached.
      return 0;
    }
    return 10;
  }
#endif

#ifdef MOTOR_TIMER_4
  int8_t
  stepMotor4Status ()
  {
    #ifdef A4988_4_ENABLE
      if (outputPinReadDigital (A4988_4_ENABLE) == LOW) return -1;
    #endif
    #ifdef A4988_4_SLEEP
      if (outputPinReadDigital (A4988_4_SLEEP) == LOW) return -2;
    #endif
    if (   (TIMSK4 & (1 << TOIE4))                                      //Timer interrupt enabled.
        && (TCCR4B & (  (1 << CS42)                                     //Timer has active clock.
                      | (1 << CS41)
                      | (1 << CS40))))
    {
      if (motorEndPos & MOTOR4POSLO) return 1;                          //Position reached.
      return 0;
    }
    return 10;
  }
#endif

#ifdef MOTOR_TIMER_5
  int8_t
  stepMotor5Status ()
  {
    #ifdef A4988_5_ENABLE
      if (outputPinReadDigital (A4988_5_ENABLE) == LOW) return -1;
    #endif
    #ifdef A4988_5_SLEEP
      if (outputPinReadDigital (A4988_5_SLEEP) == LOW) return -2;
    #endif
    if (   (TIMSK5 & (1 << TOIE5))                                      //Timer interrupt enabled.
        && (TCCR5B & (  (1 << CS52)                                     //Timer has active clock.
                      | (1 << CS51)
                      | (1 << CS50))))
    {
      if (motorEndPos & MOTOR5POSLO) return 1;                          //Position reached.
      return 0;
    }
    return 10;
  }
#endif
